import sys
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv()

from aiida.orm import load_node
from aiida.common.exceptions import InputValidationError,ValidationError
from aiida.orm.data.upf import get_pseudos_from_structure
from aiida.common.datastructures import calc_states
from collections import defaultdict
from aiida.orm.utils import DataFactory, CalculationFactory
import itertools
from aiida.orm.data.base import Float, Str, NumericType, BaseType, Bool
from aiida.work.workchain import WorkChain, while_
from aiida.work.workchain import ToContext as ResultToContext
from aiida.work.run import legacy_workflow
from aiida.work.run import run, submit
from aiida.common.links import LinkType
from aiida_yambo.workflows.yambo_utils import default_step_size, update_parameter_field, set_default_qp_param,\
               default_pw_settings, set_default_pw_param, yambo_default_settings, default_qpkrange, p2y_default_settings,\
               default_bands
from aiida.orm.data.remote import RemoteData 
from aiida.orm.data.array.kpoints import KpointsData
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida_yambo.workflows.yamborestart  import YamboRestartWf
from aiida_yambo.workflows.pwplaceholder  import PwRestartWf
from aiida_yambo.calculations.gw  import YamboCalculation
from aiida_quantumespresso.calculations.pw import PwCalculation
from aiida_quantumespresso.workflows.pw.base  import PwBaseWorkChain

ParameterData = DataFactory("parameter")

class YamboWorkflow(WorkChain):
    """
    """

    @classmethod
    def define(cls, spec):
        """Workfunction definition

        Keyword arguments:
        codename_pw -- PW code name (required)
        codename_p2y -- P2Y code name (required)
        codename_yambo -- Yambo code name (required)
        pseudo_family -- pseudo name (required)
        calculation_set_pw -- scheduler settings {'resources':{...}}  for PW calculation (required)
        calculation_set_p2y -- scheduler settings {'resources':{...}} for P2Y conversion (required)
        calculation_set_yambo -- scheduler settings {'resources':{...}} for Yambo calculation (required)
        settings_pw -- plugin settings for PW  (required)
        settings_p2y -- settings for P2Y { "ADDITIONAL_RETRIEVE_LIST":[], 'INITIALISE':True}  (optional)
        settings_yambo -- settings for yambo { "ADDITIONAL_RETRIEVE_LIST":[] } (optional)
        structure -- Structure (required)
        kpoint_pw -- kpoints  (option)
        gamma_pw -- Whether its a gammap point calculation(optional)
        parameters_pw -- PW SCF parameters (required)
        parameters_pw_nscf -- PW NSCF parameters (optional)
        parameters_p2y --  (required)
        parameters_yambo -- Parameters for Yambo (required)
        parent_folder -- Parent calculation (optional)
        previous_yambo_workchain -- Parent workchain (Yambo) (optional)
        to_set_qpkrange --  whether to set the QPkrange, override with defaults  (optional)
        to_set_bands -- Whether to set the bands, overide with default (optional)
        bands_groupname --  (optional)
        """
        super(YamboWorkflow, cls).define(spec)
        spec.input("codename_pw", valid_type=Str)
        spec.input("codename_p2y", valid_type=Str)
        spec.input("codename_yambo", valid_type=Str)
        spec.input("pseudo_family", valid_type=Str)
        spec.input("calculation_set_pw", valid_type=ParameterData) # custom_scheduler_commands,  resources,...
        spec.input("calculation_set_p2y", valid_type=ParameterData)
        spec.input("calculation_set_yambo", valid_type=ParameterData)
        spec.input("settings_pw", valid_type=ParameterData)
        spec.input("settings_p2y", valid_type=ParameterData)
        spec.input("settings_yambo", valid_type=ParameterData)
        spec.input("structure", valid_type=StructureData)
        spec.input("kpoint_pw", valid_type=KpointsData)
        spec.input("gamma_pw", valid_type=Bool, default=Bool(0), required=False )
        spec.input("parameters_pw", valid_type=ParameterData)
        spec.input("parameters_pw_nscf", valid_type=ParameterData,required=False)
        spec.input("parameters_p2y", valid_type=ParameterData)
        spec.input("parameters_yambo", valid_type=ParameterData)
        spec.input("parent_folder", valid_type=RemoteData,required=False)
        spec.input("previous_yambo_workchain", valid_type=Str,required=False)
        spec.input("to_set_qpkrange", valid_type=Bool,required=False, default=Bool(0) )
        spec.input("to_set_bands", valid_type=Bool,required=False, default=Bool(0) )
        spec.input("bands_groupname", valid_type=Str, required=False)
        spec.outline(
            cls.start_workflow,
            while_(cls.can_continue)(
                cls.perform_next,
            ),
            cls.report_wf
        )
        spec.dynamic_output()

    def start_workflow(self):
        """Initialize the workflow, set the parent calculation
        
        This function sets the parent, and its type, including support for starting from a workchain"""
        self.ctx.pw_wf_res = DataFactory('parameter')(dict={})
        self.ctx.yambo_res = DataFactory('parameter')(dict={})
        self.ctx.last_step_pw_wf = None
        self.ctx.last_step_kind = None
        self.ctx.can_cont  = 0
        self.ctx.yambo_pks = [] 
        self.ctx.pw_pks  = [] 
        self.ctx.done = False
        if 'parent_folder' in self.inputs.keys():
            parent_calc = self.inputs.parent_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
            if isinstance(parent_calc, YamboCalculation):
                self.ctx.last_step_kind = 'yambo' 
                self.ctx.yambo_res = DataFactory('parameter')(dict={"out": { "gw": {"yambo_pk": parent_calc.pk, "success": True }}} )
                self.report("Yambo calculation (pk {}) found in input, I will start from there.".format(parent_calc.pk ))
            elif isinstance(parent_calc, PwCalculation):
                self.ctx.last_step_kind = 'pw'
                self.ctx.pw_wf_res = None
                self.report("PW calculation (pk {}) found in input, I will start from there.".format(parent_calc.pk ))
            else:
                self.ctx.pw_wf_res = None
                self.report("No PW or Yambo calculation found in input, I will start from scratch.")
        if 'previous_yambo_workchain' in self.inputs.keys():
            self.report('WARNING: previous_yambo_workchain option should be used in DEBUG mode only!')
            wf_outs = load_node(int(str(self.inputs.previous_yambo_workchain)))
            self.ctx.pw_wf_res = wf_outs  # has both gw and pw outputs in one
            self.ctx.yambo_res = wf_outs 
            if 'scf_remote_folder' in wf_outs.get_outputs_dict().keys():
                scf_calc  = wf_outs.out.scf_remote_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
                if scf_calc.get_state() != u'FINISHED':
                    self.ctx.last_step_kind = 'pw'
                    del self.ctx['pw_wf_res']
            if 'nscf_remote_folder' in wf_outs.get_outputs_dict().keys():
                nscf_calc = wf_outs.out.scf_remote_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
                if nscf_calc.get_state() != u'FINISHED':
                    self.ctx.last_step_kind = 'pw'
                    del self.ctx['pw_wf_res']
            if 'yambo_remote_folder' in wf_outs.get_outputs_dict().keys():
                parent_calc = wf_outs.out.yambo_remote_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder']
                init_calc = parent_calc.inp.settings.get_dict().pop('INITIALISE', False) 
                if init_calc and parent_calc.get_state() == u'FINISHED' : # Finished P2Y
                    self.ctx.last_step_kind = 'yambo_p2y'
                elif init_calc == False and  parent_calc.get_state() != u'FINISHED':  #  Unfinished QP 
                    self.ctx.last_step_kind = 'yambo' 
                elif init_calc == False and  parent_calc.get_state() == u'FINISHED':  #  Finished QP?
                    self.ctx.last_step_kind = 'yambo' 
                else: # unifinished P2Y?
                    self.ctx.last_step_kind = 'pw'
             
            self.report("DEBUG: workchain {} loaded".format(self.ctx.yambo_res))

        try:
            self.ctx.bands_groupname = self.inputs.bands_groupname
            self.report("GW bands will be added to the group {}".format(self.inputs.bands_groupname))
        except AttributeError:
            self.ctx.bands_groupname = None
        self.ctx.parameters_yambo = self.inputs.parameters_yambo 

        self.report(" workflow initilization step completed.")
        
    def can_continue(self):
        """This function checks the status of the last calculation and determines what happens next, including a successful exit"""

        if self.ctx.last_step_kind == 'yambo' and self.ctx.yambo_res:
            try:
                self.ctx.yambo_pks.append(self.ctx.yambo_res.out.gw.get_dict()["yambo_pk"])
            except AttributeError:
                raise InputValidationError("Yambo input must be a workchain!")
            if self.ctx.yambo_res.out.gw.get_dict()["success"] == True:
                self.ctx.done = True
                self.report("Last Yambo calculation was succesful.")

        if self.ctx.last_step_kind == 'yambo_p2y' and self.ctx.yambo_res : 
            self.ctx.yambo_pks.append(self.ctx.yambo_res.out.gw.get_dict()["yambo_pk"]) 

        if self.ctx.last_step_kind == 'pw' and self.ctx.pw_wf_res != None : 
            self.ctx.pw_pks.append(self.ctx.pw_wf_res.out.pw.get_dict()["nscf_pk"]) 

        if self.ctx.done == True:
            self.report("Workflow has finished. will report outputs")
            return False
        self.ctx.can_cont +=1
        if self.ctx.can_cont> 10 :
            return False 
        return True
 
    def perform_next(self):
        """This function  will run the next step, depending on the information provided in the context"""
        if self.ctx.last_step_kind == 'yambo' or self.ctx.last_step_kind == 'yambo_p2y' :
            if load_node(self.ctx.yambo_res.out.gw.get_dict()["yambo_pk"]).get_state() == u'FINISHED':
                if self.inputs.to_set_qpkrange   and 'QPkrange' not in self.ctx.parameters_yambo.get_dict().keys():
                    self.ctx.parameters_yambo = default_qpkrange( self.ctx.pw_wf_res.out.pw.get_dict()["nscf_pk"], self.ctx.parameters_yambo)
                if self.inputs.to_set_bands   and ('BndsRnXp' not in self.ctx.parameters_yambo.get_dict().keys()\
                        or 'GbndRnge' not in self.ctx.parameters_yambo.get_dict().keys()):
                    self.ctx.parameters_yambo = default_bands( self.ctx.pw_wf_res.out.pw.get_dict()["nscf_pk"], self.ctx.parameters_yambo)
                start_from_initialize = load_node(self.ctx.yambo_res.out.gw.get_dict()["yambo_pk"]).inp.settings.get_dict().pop('INITIALISE', None)

                if start_from_initialize: # YamboRestartWf will initialize before starting QP calc  for us,  INIT != P2Y 
                    self.report ("YamboRestartWf will start from initialise mode (yambo init) ")
                    yambo_result =  self.run_yambo()
                    return  ResultToContext( yambo_res= yambo_result  )
                else:  # Possibly a restart,  after some type of failure, why was is not handled by YamboRestartWf? maybe restarting whole workchain
                    self.report(" Restarting {}, this is some form of restart for the workchain".format(self.ctx.last_step_kind)  )
                    yambo_result =  self.run_yambo()
                    return  ResultToContext( yambo_res= yambo_result  )
            if len(self.ctx.yambo_pks) > 0:
                 if load_node(self.ctx.yambo_pks[-1] ).get_state() == u'FAILED':  # Needs a resubmit depending on the error.
                    self.report("Last {} calculation (pk: {}) failed, will attempt a restart".format(self.ctx.last_step_kind, self.ctx.yambo_pks[-1] ))

        if  self.ctx.last_step_kind == 'pw' and  self.ctx.pw_wf_res :
            if self.ctx.pw_wf_res.out.pw.get_dict()['success'] == True:
                self.report("PwRestartWf was successful,  running initialize next with: YamboRestartWf ")
                p2y_result = self.run_p2y()
                return  ResultToContext( yambo_res= p2y_result )
            if self.ctx.pw_wf_res.out.pw.get_dict()['success'] == False:
                self.report("PwRestartWf subworkflow  NOT  successful")
                return 

        if  self.ctx.last_step_kind == None or self.ctx.last_step_kind == 'pw' and not self.ctx.pw_wf_res :# this is likely  the very begining, we can start with the scf/nscf here
            extra = {}
            self.report("Launching PwRestartWf ")
            if 'parameters_pw_nscf' in self.inputs.keys():
                extra['parameters_nscf'] = self.inputs.parameters_pw_nscf 
            if 'parent_folder' in self.inputs.keys():
                extra['parent_folder'] = self.inputs.parent_folder
            pw_wf_result = self.run_pw(extra)
            return  ResultToContext( pw_wf_res = pw_wf_result  )

    def run_yambo(self):
        """ submit a yambo calculation """    
        parentcalc = load_node(self.ctx.yambo_res.out.gw.get_dict()["yambo_pk"])
        parent_folder = parentcalc.out.remote_folder 
        yambo_result = submit (YamboRestartWf,precode= self.inputs.codename_p2y, yambocode=self.inputs.codename_yambo,
             parameters = self.ctx.parameters_yambo, calculation_set= self.inputs.calculation_set_yambo,
            parent_folder = parent_folder, settings = self.inputs.settings_yambo )
        self.ctx.last_step_kind = 'yambo'
        self.report ("submitted YamboRestartWf subworkflow, in Initialize mode  ")
        return yambo_result

    def run_p2y(self):
        """ submit a  P2Y  calculation """    
        parentcalc = load_node(self.ctx.pw_wf_res.out.pw.get_dict()["nscf_pk"])
        parent_folder = parentcalc.out.remote_folder 
        p2y_result = submit (YamboRestartWf, precode= self.inputs.codename_p2y, yambocode=self.inputs.codename_yambo,
             parameters = self.inputs.parameters_p2y , calculation_set= self.inputs.calculation_set_p2y,
            parent_folder = parent_folder, settings = self.inputs.settings_p2y )
        self.ctx.last_step_kind = 'yambo_p2y'
        return p2y_result

    def run_pw(self, extra):
        """ submit a  PW   calculation """    
        pw_wf_result = submit(PwRestartWf, codename = self.inputs.codename_pw  , pseudo_family = self.inputs.pseudo_family, 
                calculation_set = self.inputs.calculation_set_pw, settings=self.inputs.settings_pw, 
                kpoints=self.inputs.kpoint_pw, gamma= self.inputs.gamma_pw,
                structure = self.inputs.structure , parameters = self.inputs.parameters_pw, **extra)
        self.ctx.last_step_kind = 'pw'                   
        return pw_wf_result  

    def run_restart(self):
        """ submit a followup yambo calculation """    
        parent_folder = self.ctx.yambo_res.out.yambo_remote_folder 
        yambo_result = submit (YamboRestartWf,precode= self.inputs.codename_p2y, yambocode=self.inputs.codename_yambo,
             parameters = self.ctx.parameters_yambo, calculation_set= self.inputs.calculation_set_yambo,
            parent_folder = parent_folder, settings = self.inputs.settings_yambo )
        self.ctx.last_step_kind = 'yambo'
        self.report ("submitted YamboRestartWf subworkflow, in Initialize mode  ")
        return yambo_result

        

    def report_wf(self):
        """
        """
        self.report('Final step.')
        from aiida.orm import DataFactory
        try:
            pw = self.ctx.pw_wf_res.out.pw.get_dict()
        except Exception:
            pw = {}
        gw = self.ctx.yambo_res.out.gw.get_dict()
        #gw.update(pw)
        self.out("yambo_remote_folder", self.ctx.yambo_res.out.yambo_remote_folder)
        self.out("scf_remote_folder", self.ctx.pw_wf_res.out.scf_remote_folder)
        self.out("nscf_remote_folder", self.ctx.pw_wf_res.out.nscf_remote_folder)
        if self.ctx.bands_groupname is not None:
            g_bands,_ = Group.get_or_create(name=self.ctx.bands_groupname)
            g_bands.add_nodes(self.ctx.yambo_res)
            self.report("Yambo calc (pk: {}) added to the group {}".format(yambo_res.pk, self.ctx.bands_groupname))
        else:
            self.report ("Yambo calc done (pk: {}  ".format(gw))
        self.out("gw", DataFactory('parameter')(dict=gw ))
        self.out("pw", DataFactory('parameter')(dict=pw ))
        self.report("workflow complete")

if __name__ == "__main__":
    pass



#!/usr/bin/env python
import re
from setuptools import setup, find_packages
import json

if __name__ == '__main__':
    with open('setup.json', 'r') as info:
        kwargs = json.load(info)
    setup(
        include_package_data=True,
        setup_requires=[
            'reentry'
        ],
        reentry_register=True,
        packages=find_packages(where='.', exclude=( "gw.*", "parser.*","gw*","parser*")),
        **kwargs
    )



import sys
import copy
from aiida.backends.utils import load_dbenv, is_dbenv_loaded
import math 

if not is_dbenv_loaded():
    load_dbenv()

from aiida.common.exceptions import InputValidationError,ValidationError, WorkflowInputValidationError
from aiida.orm import load_node
from aiida.orm.data.upf import get_pseudos_from_structure
from collections import defaultdict
from aiida.orm.utils import DataFactory, CalculationFactory
from aiida.orm.code import Code
from aiida.orm.data.structure import StructureData
from aiida.common.links import LinkType
from aiida_quantumespresso.calculations.pw import PwCalculation
from aiida_yambo.calculations.gw  import YamboCalculation

ParameterData = DataFactory("parameter")

def generate_yambo_input_params(precodename,yambocodename, parent_folder, parameters,  calculation_set, settings):
    """Generate yambo plugin  inputs._*  from  given params"""
    inputs = YamboCalculation.process().get_inputs_template()
    inputs.preprocessing_code = Code.get_from_string(precodename.value)
    inputs.code = Code.get_from_string(yambocodename.value)
    calculation_set = calculation_set.get_dict()
    resource = calculation_set.pop('resources', {})
    if resource:
        inputs._options.resources =  resource 
    inputs._options.max_wallclock_seconds =  calculation_set.pop('max_wallclock_seconds', 86400) 
    max_memory_kb = calculation_set.pop('max_memory_kb',None)
    if max_memory_kb:
        inputs._options.max_memory_kb = max_memory_kb
    queue_name = calculation_set.pop('queue_name',None)
    if queue_name:
        inputs._options.queue_name = queue_name 
    custom_scheduler_commands = calculation_set.pop('custom_scheduler_commands',None)
    if custom_scheduler_commands:
        inputs._options.custom_scheduler_commands = custom_scheduler_commands
    environment_variables = calculation_set.pop("environment_variables",None)
    if environment_variables:
        inputs._options.environment_variables = environment_variables
    label = calculation_set.pop('label',None)
    if label:
        inputs._label = label 
    inputs.parent_folder = parent_folder
    inputs.settings =  settings 
    # Get defaults:
    edit_parameters = parameters.get_dict()
    try:
        calc = parent_folder.get_inputs_dict(link_type=LinkType.CREATE)['remote_folder'].inp.parent_calc_folder.get_inputs_dict()\
               ['remote_folder'].inp.parent_calc_folder.get_inputs_dict()['remote_folder']
    except AttributeError:
        calc = None
    is_pw = False
    if isinstance(calc,PwCalculation):
        is_pw = True
        nelec = calc.out.output_parameters.get_dict()['number_of_electrons']
        nbands = calc.out.output_parameters.get_dict()['number_of_bands']
        nocc = None
        if calc.out.output_parameters.get_dict()['lsda']== True or\
           calc.out.output_parameters.get_dict()['non_colinear_calculation'] == True or nbands < nelec:
           nocc = nelec/2 
        else:
           nocc = nelec/2
        #ngsblxpp = int(calc.out.output_parameters.get_dict()['wfc_cutoff']* 0.073498645/4 * 0.25)   # ev to ry then 1/4 
        ngsblxpp =  2
        nkpts = calc.out.output_parameters.get_dict()['number_of_k_points']
        if not resource:
             resource = {"num_mpiprocs_per_machine": 8, "num_machines": 1} # safe trivial defaults
        tot_mpi =  resource[u'num_mpiprocs_per_machine'] * resource[u'num_machines']
        if 'FFTGvecs' not in edit_parameters.keys():
             edit_parameters['FFTGvecs'] =  50
             edit_parameters['FFTGvecs_units'] =  'Ry'
        if 'BndsRnXp' not in edit_parameters.keys():
             edit_parameters['BndsRnXp'] = ( 1.0 , nocc *12 )
        if 'GbndRnge' not in edit_parameters.keys():
             edit_parameters['GbndRnge'] = (1.0,  nocc  *12 ) 
        if 'NGsBlkXp' not in edit_parameters.keys():
             edit_parameters['NGsBlkXp'] = ngsblxpp
             edit_parameters['NGsBlkXp_units'] =  'Ry'
        if 'QpntsRXp' not in edit_parameters.keys():
             edit_parameters['QpntsRXp'] = (1, 301 ) 
        if 'rim_cut' not in edit_parameters.keys():
             edit_parameters['rim_cut'] = True
        if 'GTermEn' not in edit_parameters.keys():
             edit_parameters['GTermEn'] = 40.81708
             edit_parameters['GTermEn_units'] = 'eV'
             edit_parameters['GTermKind'] = 'BG'
        if 'BoseTemp' not in edit_parameters.keys():
             edit_parameters['BoseTemp'] = 0.000
             edit_parameters['BoseTemp_units'] = 'eV'
        if 'ElecTemp' not in edit_parameters.keys():
             edit_parameters['ElecTemp'] = 0.000
             edit_parameters['ElecTemp_units'] = 'eV'
        if 'RandQpts' not in edit_parameters.keys():
             edit_parameters['RandQpts'] = 1000000
        if 'RandGvec' not in edit_parameters.keys():
             edit_parameters['RandGvec'] = 1
             edit_parameters['RandGvec_units'] = 'RL'
        if 'CUTGeo' not in edit_parameters.keys():
             edit_parameters['CUTGeo'] = 'none'
        if 'QPkrange' not in edit_parameters.keys():
             edit_parameters['QPkrange'] = [(1,1,int(nocc)-1, int(nocc)+1 ), (7,7,int(nocc)-1, int(nocc)+1 )] # To revisit 
        if 'SE_CPU' not in  edit_parameters.keys():
            qp, b = split_incom(tot_mpi)
            edit_parameters['SE_CPU'] ="1 {qp} {b}".format(qp=qp, b = b) 
            edit_parameters['SE_ROLEs']= "q qp b"
        if 'X_all_q_CPU' not in  edit_parameters.keys():
            v, c = split_incom(tot_mpi)
            edit_parameters['X_all_q_CPU']= "1 1 {c} {v}".format(c = c, v = v)
            edit_parameters['X_all_q_ROLEs'] ="q k c v"
    inputs.parameters = ParameterData(dict=edit_parameters) 
    return  inputs

def get_pseudo(structure, pseudo_family):
    """Get pseudopotential by family for atoms ina  structure """
    kind_pseudo_dict = get_pseudos_from_structure(structure, pseudo_family)
    pseudo_dict = {}
    pseudo_species = defaultdict(list)
    for kindname, pseudo in kind_pseudo_dict.iteritems():
        pseudo_dict[pseudo.pk] = pseudo
        pseudo_species[pseudo.pk].append(kindname)
    pseudos = {}
    for pseudo_pk in pseudo_dict:
        pseudo = pseudo_dict[pseudo_pk]
        kinds = pseudo_species[pseudo_pk]
        for kind in kinds:
            pseudos[kind] = pseudo
    return pseudos


def generate_pw_input_params(structure, codename, pseudo_family,parameters, calculation_set, kpoints,gamma,settings,parent_folder):
    """Generate PW input._*  from given params """
    inputs = {}
    inputs['structure'] = structure
    inputs['code'] = Code.get_from_string(codename.value)
    calculation_set = calculation_set.get_dict() 
    inputs['options'] = ParameterData(dict=calculation_set)
    if parent_folder:
        inputs['parent_folder'] = parent_folder
    inputs['kpoints']=kpoints
    inputs['parameters'] = parameters  
    inputs['pseudo_family'] =  pseudo_family
    inputs['settings']  = settings
    return  inputs


def reduce_parallelism(typ, roles,  values,calc_set):
    """
       calculation_set_yambo ={'resources':  {"num_machines": 8,"num_mpiprocs_per_machine": 32}, 'max_wallclock_seconds': 200,
            'max_memory_kb': 1*92*1000000 ,  'custom_scheduler_commands': u"#PBS -A  Pra14_3622" ,
            '  environment_variables': {"OMP_NUM_THREADS": "2" }  
            }
    """
    calculation_set = copy.deepcopy(calc_set)
    # the latter needs to be reduced, we can either increase the former or leave it untouched.
    # lets reduce it by  50% if its >=2, else increase num_machines, holding it constant at 1
    num_machines = calculation_set['resources']['num_machines']    
    num_mpiprocs_per_machine = calculation_set['resources']['num_mpiprocs_per_machine']
    omp_threads=1
    if 'environment_variables' in calculation_set.keys():
        omp_threads = calculation_set['environment_variables'].pop('OMP_NUM_THREADS',1)
    num_mpiprocs_per_machine=int(num_mpiprocs_per_machine/2)
    omp_threads= int(omp_threads)*2
    increased_mpi_task = False 
    if num_mpiprocs_per_machine < 1:
        num_mpiprocs_per_machine = 1 
        num_machines = num_machines * 2
        increased_mpi_task = True 
    calculation_set['environment_variables']['OMP_NUM_THREADS'] = str(omp_threads)
    calculation_set['environment_variables']['NUM_CORES_PER_MPIPROC'] = str(omp_threads)
    if isinstance(values,list):
        values = values[0]
    if isinstance(roles,list):
        roles = roles[0]
    # adjust the X_all_q_CPU and SE_CPU
    mpi_task = num_mpiprocs_per_machine*num_machines 
    if typ == 'X_all_q_CPU':
        #X_all_q_CPU = "1 1 96 32"
        #X_all_q_ROLEs = "q k c v"
        X_para = [ int(it) for it in values.strip().split(' ') if it ]
        try:
            c_index = roles.split(' ').index("c")
            v_index = roles.split(' ').index("v")
            c = X_para[c_index] or 1
            v = X_para[v_index] or 1
        except ValueError:
            c_index = v_index = 0
            c = 1
            v = 1
        except IndexError:
            c_index = v_index = 0
            c = 1
            v = 1
        # we should keep c*v == mpi_task, with  c>v always
        # if increased_mpi_task we try a simple assignment
        # else, we factor again?  
        v, c = split_incom(mpi_task*2)
        if  False:
            if num_mpiprocs_per_machine < calculation_set['resources']['num_mpiprocs_per_machine'] :
                c = c/2 
            if num_machines > calculation_set['resources']['num_machines']:
                c = c*2 
        if False: 
            if num_mpiprocs_per_machine < calculation_set['resources']['num_mpiprocs_per_machine'] and v >1:
                v = v/2 
            if num_machines > calculation_set['resources']['num_machines']:
                c = c*2 

        if c_index and v_index:
            X_para[c_index] = c  
            X_para[v_index] = v
        X_string = " ".join([str(it) for it in X_para])
        if X_string.strip() == '':
            mpi_per = calculation_set['resources']['num_mpiprocs_per_machine'] 
            num_mach = calculation_set['resources']['num_machines']
            c = mpi_per if mpi_per > num_mach else num_mach
            v = mpi_per if mpi_per < num_mach else num_mach
            X_string = '1  1 {c} {v}'.format(c=c, v=v )
        calculation_set['resources']['num_machines'] = num_machines
        calculation_set['resources']['num_mpiprocs_per_machine'] = num_mpiprocs_per_machine
        return X_string , calculation_set
            
    if typ == 'SE_CPU':
        #SE_ROLEs = "q qp b"
        #SE_CPU = "1 32 96"
        SE_para  = [ int(it) for it in values.strip().split(' ') if it ]
        try:
            qp_index = roles.split(' ').index("qp")
            b_index = roles.split(' ').index("b")
            qp = SE_para[qp_index] or  1
            b  = SE_para[b_index] or 1  
        except ValueError:
            qp_index = b_index = 0
            qp =1
            b  =1  
        except IndexError:
            qp_index = b_index = 0
            qp =1
            b  =1  
        if num_mpiprocs_per_machine < calculation_set['resources']['num_mpiprocs_per_machine'] and qp >1:
             qp = qp/2 
        if num_mpiprocs_per_machine < calculation_set['resources']['num_mpiprocs_per_machine'] and qp == 1:
            qp = qp/2 
        if num_machines > calculation_set['resources']['num_machines']: 
            b = b*2 
        if qp_index and b_index: 
            SE_para[qp_index] = qp  
            SE_para[b_index] = b
        SE_string = " ".join([str(it) for it in SE_para])
        if SE_string.strip() == '':
            mpi_per = calculation_set['resources']['num_mpiprocs_per_machine']
            num_mach = calculation_set['resources']['num_machines']
            c = mpi_per if mpi_per > num_mach else num_mach
            v = mpi_per if mpi_per < num_mach else num_mach
            SE_string = '1   {v} {c}'.format(c=c, v=v )
        
        calculation_set['resources']['num_machines'] = num_machines
        calculation_set['resources']['num_mpiprocs_per_machine'] = num_mpiprocs_per_machine
        return SE_string, calculation_set
   

default_step_size = {
               'PPAPntXp': .2,   
               'NGsBlkXp': .1,   
               'BSENGBlk': .1,  
               'BSENGexx': .2,
               'BndsRnXp': .1, 
               'GbndRnge': .1,
               'BSEBands': 2, # 
               'FFTGvecs': .2,
                }

def update_parameter_field( field, starting_point, update_delta):
    """Update a parameter in the parameter dictionary given the starting point and the update to apply"""
    if field in ['PPAPntXp','NGsBlkXp','BSENGBlk','BSENGexx','FFTGvecs']: # single numbers
        new_field_value =  starting_point  + update_delta 
        return new_field_value
    elif field == 'BndsRnXp':
        new_hi_value =  int( starting_point  + update_delta )
        new_low_value =  int( starting_point  - update_delta )
        #return (new_low_value , new_hi_value)
        return ( 1 , new_hi_value)
    elif field == 'GbndRnge':
        new_field_value =  int( starting_point   + update_delta )
        return (1, new_field_value)
    elif field == 'BSEBands':  # Will be useful when we support BSE calculations
        hi =  starting_point +   update_delta
        low  =  starting_point  -  update_delta
        return ( low, hi )
    else:
        raise WorkflowInputValidationError("convergences the field {} are not supported".format(field))


def set_default_qp_param(parameter=None):
    """Create a default QP parameter object"""
    if not parameter:
       parameter = ParameterData(dict={})
    edit_param = parameter.get_dict()
    if 'ppa' not in edit_param.keys():
        edit_param['ppa'] = True
    if 'gw0' not in edit_param.keys():
        edit_param['gw0'] = True
    if 'HF_and_locXC' not in edit_param.keys():
        edit_param['HF_and_locXC'] = True
    if 'em1d' not in edit_param.keys():
        edit_param['em1d'] = True
    if 'DysSolver' not in edit_param.keys():
        edit_param['DysSolver'] = "n"
    if 'Chimod' not in edit_param.keys():
        edit_param['Chimod'] = "Hartree"
    if 'LongDrXp' not in edit_param.keys():
        edit_param['LongDrXp'] = (1.000000,0.000000, 0.000000)
    if 'PPAPntXp' not in edit_param.keys():
        edit_param['PPAPntXp'] =  20
        edit_param['PPAPntXp_units'] =  'eV'
    if 'QpntsRXp' not in edit_param.keys():
         edit_param['QpntsRXp'] = (1, 301 ) 
    if 'rim_cut' not in edit_param.keys():
         edit_param['rim_cut'] = True
    if 'GTermEn' not in edit_param.keys():
         edit_param['GTermEn'] = 40.81708
         edit_param['GTermEn_units'] = 'eV'
         edit_param['GTermKind'] = 'BG'
    if 'BoseTemp' not in edit_param.keys():
         edit_param['BoseTemp'] = 0.000
         edit_param['BoseTemp_units'] = 'eV'
    if 'ElecTemp' not in edit_param.keys():
         edit_param['ElecTemp'] = 0.000
         edit_param['ElecTemp_units'] = 'eV'
    if 'RandQpts' not in edit_param.keys():
         edit_param['RandQpts'] = 1000000
    if 'RandGvec' not in edit_param.keys():
         edit_param['RandGvec'] = 1
         edit_param['RandGvec_units'] = 'RL'
    if 'CUTGeo' not in edit_param.keys():
         edit_param['CUTGeo'] = 'none'
    if 'SE_CPU' not in  edit_param.keys():
        edit_param['SE_CPU'] ="1 8 16" 
        edit_param['SE_ROLEs']= "q qp b"
    if 'X_all_q_CPU' not in  edit_param.keys():
        edit_param['X_all_q_CPU']= "1 1 16 8"
        edit_param['X_all_q_ROLEs'] ="q k c v"
    if 'FFTGvecs' not in edit_param.keys():
        edit_param['FFTGvecs'] =  50
        edit_param['FFTGvecs_units'] =  'Ry'
    if 'NGsBlkXp' not in edit_param.keys():
        edit_param['NGsBlkXp'] =  2
        edit_param['NGsBlkXp_units'] =  'Ry'
    return ParameterData(dict=edit_param)



def set_default_pw_param(nscf=False):
    """Create a default PW  parameter object"""
    pw_parameters =  {
          'CONTROL': {
              'calculation': 'scf',
              'restart_mode': 'from_scratch',
              'wf_collect': True,
              'verbosity' :'high',
              },
          'SYSTEM': {
              'ecutwfc': 70.,
              'occupations':'smearing',
              'degauss': 0.0001,
              'starting_magnetization(2)' : 0.1,
              'smearing': 'fermi-dirac',
              },
          'ELECTRONS': {
              'conv_thr': 1.e-6,
              'electron_maxstep ': 100,
              } }
    if nscf:
        pw_parameters['CONTROL']['calculation'] = 'nscf'
        pw_parameters['SYSTEM']['force_symmorphic'] =  True 
    return ParameterData(dict=pw_parameters)

def default_pw_settings():
    return ParameterData(dict={})
 
def yambo_default_settings():
    return ParameterData(dict={"ADDITIONAL_RETRIEVE_LIST":[
                             'r-*','o-*','l-*','l_*','LOG/l-*_CPU_1','aiida/ndb.QP','aiida/ndb.HF_and_locXC'] })      
 
def p2y_default_settings():
    return ParameterData(dict={ "ADDITIONAL_RETRIEVE_LIST":['r-*','o-*','l-*','l_*','LOG/l-*_CPU_1'], 'INITIALISE':True})


def default_qpkrange(calc_pk, parameters):
    """ Create a default QPkrange when one is not provided"""
    calc = load_node(calc_pk)
    edit_parameters = parameters.get_dict()
    if isinstance(calc,PwCalculation):
       nelec = calc.out.output_parameters.get_dict()['number_of_electrons']
       nbands = calc.out.output_parameters.get_dict()['number_of_bands']
       nocc = None
       if calc.out.output_parameters.get_dict()['lsda']== True or\
               calc.out.output_parameters.get_dict()['non_colinear_calculation'] == True or nbands < nelec:
          nocc = nelec/2 
       else:
          nocc = nelec/2
       is_pw = True
       nkpts = calc.out.output_parameters.get_dict()['number_of_k_points']
       if 'QPkrange' not in edit_parameters.keys():
            edit_parameters['QPkrange'] = [(1, 1, int(nocc) , int(nocc)+1 ), (7,7,int(nocc)-1, int(nocc)+1 )]
    return ParameterData(dict=edit_parameters)


def default_bands(calc_pk, parameters):
    """Set default range for the chi bands and G  bands"""
    calc = load_node(calc_pk)
    edit_parameters = parameters.get_dict()
    if isinstance(calc,PwCalculation):
       nelec = calc.out.output_parameters.get_dict()['number_of_electrons']
       nbands = calc.out.output_parameters.get_dict()['number_of_bands']
       nocc = None
       if calc.out.output_parameters.get_dict()['lsda']== True or\
               calc.out.output_parameters.get_dict()['non_colinear_calculation'] == True or nbands < nelec:
          nocc = nelec/2 
       else:
          nocc = nelec/2
       edit_parameters['BndsRnXp'] = ( 1.0 , nocc *12)
       edit_parameters['GbndRnge'] = (1.0,  nocc  *12)
    return ParameterData(dict=edit_parameters)

def split_incom(num):
    """Does not work yet, needs some bugfix. 

    so when the power p is odd, take (p-1)/2 and (p-1)/2+1
    when the power p is even, take (p/2)+1 and (p/2)-1
    let's do an example, random number 563. largest power of 2 still less than is 9
    so we use the formula for odd p and take (p-1)/2 and (p-1)/2+1 which 4 and 5
    so 16 and 32
    """
    powers = []
    i = 1
    while i <= num:
        if i & num:
            powers.append(i)
        i <<= 1
    larges_p = powers[-1]
    power = int(math.log(larges_p)/math.log(2))
    if power%2 == 0:
        return (2**(power/2+1), 2**(power/2-1))
    else:
        return (2**((power-1)/2+1), 2**((power-1)/2))

def is_converged(values,conv_tol=1e-5,conv_window=3):
    """Check convergence for a list of values
    
    If the change between successive iterations is less than conv_tol for conv_window iterations
    the list is said to be converged.

    :param values: list of values in input
    :param conv_tol: convergence tolerance (optional, default = 1e-5)
    :param conv_window: number of last iterations considered (optional, default = 3)
    :rtype: Bool
    """
    delta_list = []
    for i in range(1,len(values)):
        delta_list.append(abs(values[i]-values[i-1]))
    delta_list = delta_list[-conv_window:]
    return all(x<conv_tol for x in delta_list) 


def default_convergence_settings():
    return ParameterData(dict={
          'start_fft' :20 , 
          'max_fft': 400 , 
          'start_bands' : 10 ,
          'max_bands': 4000, 
          'start_w_cutoff': 1 ,
          'max_w_cutoff': 40 ,
          'kpoint_starting_distance': 1.,
          'kpoint_min_distance': 0.025}) 
